'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _stringify = require('babel-runtime/core-js/json/stringify');

var _stringify2 = _interopRequireDefault(_stringify);

var _tokenStream = require('token-stream');

var _tokenStream2 = _interopRequireDefault(_tokenStream);

var _babelCodeFrame = require('babel-code-frame');

var _babelCodeFrame2 = _interopRequireDefault(_babelCodeFrame);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function parse(tokens, file) {
  tokens = new _tokenStream2.default(tokens);
  function error(msg, tokenOrNode) {
    throw new Error(msg + '\n\n' + (0, _babelCodeFrame2.default)(file.code, tokenOrNode.loc.start.line, tokenOrNode.loc.start.column));
  }
  function assert(value, msg, token) {
    if (!value) {
      throw error(msg, token);
    }
  }
  function parseArg() {
    var identifier = tokens.advance();
    assert(identifier.type === 'Identifier', 'Expected an identifier', identifier);

    var colon = tokens.advance();
    assert(colon.type === 'Colon', 'Expected a colon', colon);

    var val = tokens.advance();
    assert(val.type === 'Expression' || val.type === 'String' || val.type === 'Bool' || val.type === 'Number', 'Expected an expression, string, boolean or number', val);

    return { type: 'Arg', name: identifier.val, val: val, loc: { start: identifier.loc.start, end: identifier.loc.end } };
  }
  function parseArgs() {
    var openBracket = tokens.advance();
    assert(openBracket.type === 'Bracket' && openBracket.val === '(', 'Expected an open round brackets', openBracket);
    var args = [];

    var _loop = function _loop() {
      var arg = parseArg();
      if (args.some(function (a) {
        return a.name === arg.name;
      })) {
        throw error('Duplicate argument with name ' + (0, _stringify2.default)(arg.name), arg);
      }
      args.push(arg);
    };

    while (tokens.peek().type !== 'Bracket') {
      _loop();
    }
    var closeBracket = tokens.advance();
    assert(closeBracket.type === 'Bracket' && closeBracket.val === ')', 'Expected a close round brackets', closeBracket);
    return args;
  }
  function parseFieldSet() {
    var openBracket = tokens.advance();
    assert(openBracket.type === 'Bracket' && openBracket.val === '{', 'Expected open curly bracket', openBracket);
    var body = parseBody();
    var closeBracket = tokens.advance();
    assert(closeBracket.type === 'Bracket' && closeBracket.val === '}', 'Expected close curly bracket', closeBracket);
    return { type: 'FieldSet', body: body, loc: { start: openBracket.loc.start, end: closeBracket.loc.end } };
  }
  function parseAlias() {
    var keyword = tokens.advance();
    assert(keyword.type === 'Keyword' && keyword.val === 'as', 'Expected "as" keyword', keyword);
    var identifier = tokens.advance();
    assert(identifier.type === 'Identifier', 'Expected identifier', identifier);
    return identifier.val;
  }
  function parseField() {
    var identifier = tokens.advance();
    assert(identifier.type === 'Identifier', 'Expected identifier', identifier);
    var end = identifier.loc.end;
    var args = null;
    var body = null;
    var alias = null;
    if (tokens.peek().type === 'Bracket' && tokens.peek().val === '(') {
      args = parseArgs();
      if (args.length) {
        end = args[args.length - 1].loc.end;
      }
    }
    if (tokens.peek().type === 'Keyword' && tokens.peek().val === 'as') {
      end = tokens.peek().loc.end;
      alias = parseAlias();
      if (alias === identifier.val) alias = null;
    }
    if (tokens.peek().type === 'Bracket' && tokens.peek().val === '{') {
      body = parseFieldSet();
      end = body.loc.end;
    }
    return { type: 'Field', name: identifier.val, args: args, alias: alias, body: body, loc: { start: identifier.loc.start, end: end } };
  }
  function parseSpread() {
    var spread = tokens.advance();
    assert(spread.type === 'Spread', 'Expected spread element', spread);
    var arg = tokens.advance();
    assert(arg.type === 'Expression', 'Expected expression', arg);
    return { type: 'Spread', arg: arg, loc: { start: spread.loc.start, end: arg.loc.end } };
  }
  function parseKeyword() {
    var keyword = tokens.advance();
    assert(keyword.type === 'Keyword' && keyword.val === 'if', 'Expected "if" keyword', keyword);
    var condition = tokens.advance();
    assert(condition.type === 'Expression', 'Expected expression', condition);
    var consequent = parseFieldSet();
    var alternate = null;
    if (tokens.peek().type !== 'Keyword' && tokens.peek().val === 'else') {
      tokens.advance();
      alternate = parseFieldSet();
    }
    return {
      type: 'Conditional',
      condition: condition,
      consequent: consequent,
      alternate: alternate,
      loc: { start: keyword.loc.start, end: (alternate || consequent).loc.end }
    };
  }
  function parseBody() {
    var body = [];
    var usedIdentifiers = [];
    while ((tokens.peek().type !== 'Bracket' || tokens.peek().val !== '}') && tokens.peek().type !== 'End') {
      switch (tokens.peek().type) {
        case 'Identifier':
          var field = parseField();
          assert(typeof field.name === 'string', 'Field.name should be a string', field);
          assert(field.alias === null || typeof field.alias === 'string', 'Field.alias should be a string or null', field);
          var fieldName = field.alias || field.name;
          if (usedIdentifiers.indexOf(fieldName) !== -1) {
            throw error('Duplicate key ' + (0, _stringify2.default)(fieldName), field);
          }
          usedIdentifiers.push(fieldName);
          body.push(field);
          break;
        case 'Spread':
          body.push(parseSpread());
          break;
        case 'Keyword':
          body.push(parseKeyword());
          break;
        default:
          throw error('Unexpected token ' + tokens.peek().type + ', expected Identifier, Spread or Keyword', tokens.peek());
      }
    }
    return body;
  }
  var start = tokens.peek().loc.start;
  var body = parseBody();
  var eof = tokens.advance();
  assert(eof.type === 'End', 'Unexpected token ' + eof.type, eof);
  return {
    type: 'FieldSet', body: body, loc: { start: start, end: eof.loc.end }
  };
}
exports.default = parse;